; System.asm
;
; System interface support for the Bootstrap language
; Note: This code is not thread-safe.
;
; 2005-09-19
;   Added buffer overflow check to "get command line" function.
;   Added preservation of ESI and EDI, as per Win32, VC++, and Intel ABI.
;
; 2005-09-20
;   Added base register for system function arguments. It is retrieved from the stack.
;
; 2005-09-24
;   Added the use of header information for finding the user code entry point.
;   Added the allocation of data space from the heap.
;
; Release as Runtime Version 1.00

        .386
        .model  flat

ExitProcess     textequ <_ExitProcess@4>
GetCommandLine  textequ <_GetCommandLineA@0>
CreateFile      textequ <_CreateFileA@28>
CloseHandle     textequ <_CloseHandle@4>
ReadFile        textequ <_ReadFile@20>
WriteFile       textequ <_WriteFile@20>
GetProcessHeap  textequ <_GetProcessHeap@0>
HeapAlloc       textequ <_HeapAlloc@12>

; Header section offsets
hdrsize         equ     0       ; size of header section
hdrrunver       equ     4       ; run time (startup and library) version
hdrcodesize     equ     8       ; size of code section
hdrdatasize     equ     12      ; size of data section

; System function argument structure offsets
sysfn           equ     0       ; system function number
sysarg1         equ     4       ; first system function argument
sysarg2         equ     8       ; second system function argument
sysarray        equ     12      ; system function array argument

        extern  _pgm:dword      ; start of load module

        extern  ExitProcess:near
        extern  GetCommandLine:near
        extern  CreateFile:near
        extern  CloseHandle:near
        extern  ReadFile:near
        extern  WriteFile:near
        extern  GetProcessHeap:near
        extern  HeapAlloc:near

NULL                    equ     0

INVALID_HANDLE_VALUE    equ     -1

GENERIC_READ            equ     80000000h
GENERIC_WRITE           equ     40000000h

FILE_SHARE_READ         equ     1h
FILE_SHARE_WRITE        equ     2h

CREATE_ALWAYS           equ     2
OPEN_EXISTING           equ     3

MB_OK                   equ     0h
MB_ICONERROR            equ     10h
MB_TOPMOST              equ     40000h

        .code

;=======================================
; This is the starting point immediately after
; the EXE file has been loaded.
;
; When user code is called, these registers are set:
;
;   EAX = address of system function
;   ESI = address of allocated data area

_start:
        mov     ebx,offset _pgm         ; Get start of load module
        add     ebx,4                   ; Skip magic number to get start of header

        push    [ebx+hdrdatasize]       ; Size for HeapAlloc
        push    0                       ; Flags for HeapAlloc
        call    GetProcessHeap          ; (no args)
        push    eax                     ; Heap handle for HeapAlloc
        call    HeapAlloc               ;
        test    eax,eax                 ; If handle = NULL,
        jz      heap_error              ;   then handle allocation error
        mov     esi,eax                 ; Copy address of allocated storage

        add     ebx,[ebx]               ; Add header size to get start address
        mov     eax,offset system       ; Get function address
        call    ebx                     ; Start user code
heap_error:
        push    0                       ; Set system (error) return value
        call    ExitProcess             ;

;=======================================
; The entry point to the system support function.
; As per Win32, VC++, and Intel ABI, the registers
; EBX, ESI, and EDI must be preserved. The other
; general purpose registers EAX, ECX, and EDX are
; allowed to change without restoration.
;   The direction (descending) flag, DF, must be
; cleared when this function is called. This function
; must leave DF cleared before exiting.
;   Current calling convention is CDECL.

system:
        mov     ecx,[esp+4]             ; Get pointer to monitor argument structure
        mov     eax,[ecx+sysfn]         ; Get system function number
        test    eax,eax                 ;
        jz      sysfn0                  ;
        dec     eax                     ;
        jz      sysfn1                  ;
        dec     eax                     ;
        jz      sysfn2                  ;
        dec     eax                     ;
        jz      sysfn3                  ;
        dec     eax                     ;
        jz      sysfn4                  ;
        dec     eax                     ;
        jz      sysfn5                  ;

; invalid function
        ret                             ;

;=======================================
; The individual system functions follow...
;=======================================

; =====
; system function 0 - exit program
; =====
sysfn0:
        push    0                       ;
        call    ExitProcess             ;

; =====
; system function 1 - get command line
; =====
sysfn1:
        push    esi                     ;
        push    edi                     ;
        push    ebx                     ;
        mov     ebx,ecx                 ; Set new base register
        call    GetCommandLine          ;
        mov     ecx,4095                ; Counter for preventing buffer overflow
        mov     esi,eax                 ; Pointer to command line string
        mov     edi,offset TempBuffer   ;
copy_command:
        mov     al,[esi]                ;
        inc     esi                     ;
        mov     [edi],al                ;
        inc     edi                     ;
        dec     ecx                     ; Buffer full?
        jz      quit_copy_command       ; Yes
        or      al,al                   ; End of string (zero terminator byte)?
        jnz     copy_command            ;
quit_copy_command:
        mov     TempBuffer+4095,0       ; Ensure string is zero terminated within 4096 byte buffer
        call    BufferToInt             ; Unpack string
        pop     ebx                     ;
        pop     edi                     ;
        pop     esi                     ;
        ret                             ;

; =====
; system function 2 - open file
; =====
sysfn2:
        push    ebx                     ;
        mov     ebx,ecx                 ; Set new base register
        call    BufferToChar            ; Pack file name
        mov     eax,[ebx+sysarg1]       ; Get read/write flag
        test    eax,eax                 ;
        jnz     open_write              ;
open_read:
        push    NULL                    ; No template (prototype) file
        push    0                       ; No special file flags and attributes
        push    OPEN_EXISTING           ; Open type
        push    NULL                    ; Use default security (no inheritance)
        push    FILE_SHARE_READ         ; Share mode
        push    GENERIC_READ            ; I/O mode
        push    offset TempBuffer       ; File name
        call    CreateFile              ;
        mov     [ebx+sysarg1],eax       ; Store file handle
        pop     ebx                     ;
        ret                             ;

open_write:
        push    NULL                    ; No template (prototype) file
        push    0                       ; No special file flags and attributes
        push    CREATE_ALWAYS           ; Open type
        push    NULL                    ; Use default security (no inheritance)
        push    FILE_SHARE_READ         ; Share mode
        push    GENERIC_WRITE           ; I/O mode
        push    offset TempBuffer       ; File name
        call    CreateFile              ;
        mov     [ebx+sysarg1],eax       ; Store file handle
        pop     ebx                     ;
        ret                             ;

; =====
; system function 3 - close file
; =====
sysfn3:
        push    ebx                     ;
        mov     ebx,ecx                 ; Set new base register
        push    [ebx+sysarg1]           ; File handle
        call    CloseHandle             ;
        pop     ebx                     ;
        ret                             ;

; =====
; system function 4 - read one byte from file
; =====
sysfn4:
        push    ebx                     ;
        mov     ebx,ecx                 ; Set new base register
        push    NULL                    ; No overlapped I/O
        push    offset dwBytesXferred   ; Pointer to number of bytes retrieved
        push    1                       ; Number of bytes
        push    offset TempBuffer       ; Pointer to buffer that receives data
        push    [ebx+sysarg2]           ; File handle
        call    ReadFile                ;
        cmp     dwBytesXferred,0        ; Assume EOF if no bytes read
        je      signalEOF               ;
        xor     eax,eax                 ; Clear upper bits
        mov     al,TempBuffer           ;
        mov     [ebx+sysarg1],eax       ; Return a character
        pop     ebx                     ;
        ret                             ;
signalEOF:
        mov     dword ptr[ebx+sysarg1],-1 ; EOF "character"
        pop     ebx                     ;
        ret                             ;

; =====
; system function 5 - write one byte to file
; =====
sysfn5:
        push    ebx                     ;
        mov     ebx,ecx                 ; Set new base register
        mov     eax,[ebx+sysarg1]       ;
        mov     TempBuffer,al           ; Store character to write
        push    NULL                    ; No overlapped I/O
        push    offset dwBytesXferred   ; Pointer to number of bytes written
        push    1                       ; Number of bytes
        push    offset TempBuffer       ; Pointer to buffer containing data to write
        push    [ebx+sysarg2]           ; File handle
        call    WriteFile               ;
        pop     ebx                     ;
        ret                             ;

;=======================================
; Buffer conversion routines
;=======================================

; Copy sysarray to TempBuffer, packing characters
BufferToChar:
        push    esi                     ;
        push    edi                     ;
        lea     esi,[ebx+sysarray]      ;
        mov     edi,offset TempBuffer   ;
        mov     ecx,4096                ;
ToChar1:
        mov     eax,[esi]               ; Get a dword
        add     esi,4                   ;
        mov     [edi],al                ; Store a byte
        inc     edi                     ;
        dec     ecx                     ;
        jnz     ToChar1                 ;
        pop     edi                     ;
        pop     esi                     ;
        ret                             ;

; Copy TempBuffer to sysarray, unpacking characters
BufferToInt:
        push    esi                     ;
        push    edi                     ;
        xor     eax,eax                 ;
        mov     esi,offset TempBuffer   ;
        lea     edi,[ebx+sysarray]      ;
        mov     ecx,4096                ;
ToInt1:
        mov     al,[esi]                ; Get one byte
        inc     esi                     ;
        mov     [edi],eax               ; Store a whole dword
        add     edi,4                   ;
        dec     ecx                     ;
        jnz     ToInt1                  ;
        pop     edi                     ;
        pop     esi                     ;
        ret                             ;

.data
dwBytesXferred  dd      ?
TempBuffer      db      4096 dup(?)

        end     _start
